from aq_utilities.data.daily_stations.psql import daily_stations_to_postgres, load_daily_stations
from aq_utilities.data.hourly_data.psql import hourly_predictions_to_postgres
from aq_utilities.data.hourly_data.psql import hourly_features_to_postgres
from aq_utilities.data.hourly_data.psql import hourly_data_to_postgres
from aq_utilities.data.hourly_data.psql import load_hourly_data
from aq_utilities.data.hourly_data.psql import load_hourly_feature
from aq_utilities.data.stations_info.psql import stations_info_to_postgres
from aq_utilities.data.stations_info.psql import load_stations_info
from aq_utilities.data.failures.psql import write_failure_to_postgres
from aq_utilities.data.psql import get_max_timestamp, get_min_timestamp
from aq_utilities.data.processing.filter_stations import *
from aq_utilities.data.processing.reindex import *



from datetime import datetime
from typing import Tuple, Union

import pandas as pd
import psycopg2

from aq_utilities.config.data import CHUNCKSIZE


def write_failure_to_postgres(failed_request: Tuple["datetime", str],
                              target_table_name: str,
                              chunksize: int = CHUNCKSIZE,
                              engine: "sqlalchemy.engine.Engine" = None,
                              verbose: bool = False) -> None:
    """Write failure to postgres database."""
    # the schema for failures is constant (cols timestamp and request)
    df = pd.DataFrame([failed_request], columns=["timestamp", "request"])
    df.timestamp = pd.to_datetime(df.timestamp)
    try:
        if verbose:
            print(
                f"[{datetime.now()}] writing failure {failed_request} to postgres table {table_name}"
            )
        df.to_sql(
            target_table_name,
            engine,
            if_exists="append",
            index=False,
            chunksize=chunksize,
            method="multi",
        )
        if verbose: print(f"[{datetime.now()}] wrote failure")
    except Exception as e:
        print(f"failed to write failure to postgres: {e}")



import os

from sqlalchemy import create_engine

PGUSER = os.environ.get("PGUSER", "pgadmin")
PGPASSWORD = os.environ.get("PGPASSWORD", "")
PGHOST = os.environ.get("PGHOST", "localhost")
PGDATABASE = os.environ.get("PGDATABASE", "aq")
ENGINE = None


def get_engine(
    pg_user: str = PGUSER,
    pg_password: str = PGPASSWORD,
    pg_host: str = PGHOST,
    pg_database: str = PGDATABASE,
):
    global ENGINE
    if ENGINE is None:
        ENGINE = create_engine(
            f"postgresql+psycopg2://{pg_user}:{pg_password}@{pg_host}/{pg_database}",
            echo=False,
            execution_options={"isolation_level": "AUTOCOMMIT"},
            connect_args={"sslmode": "require"},
            pool_size=10,
            max_overflow=10,
            pool_timeout=30,
        )
    return ENGINE



import time
from datetime import datetime
from pathlib import Path
from typing import Union, List

import pandas as pd
import psycopg2
from sqlalchemy.sql import text

from aq_utilities.config.data import CHUNCKSIZE
from aq_utilities.data.remote import download_file
from aq_utilities.data.failures.psql import write_failure_to_postgres


def hourly_predictions_to_postgres(predictions: pd.DataFrame,
                                   engine: "sqlalchemy.engine.Engine",
                                   chunksize: int = CHUNCKSIZE,
                                   verbose: bool = False) -> int:
    """Load hourly data to postgres database."""
    if verbose:
        print(
            f"[{datetime.now()}] writing {len(predictions)} records to postgres"
        )
    try:
        predictions.to_sql(
            "hourly_predictions",
            engine,
            if_exists="append",
            index=False,
            chunksize=chunksize,
            method="multi",
        )
        if verbose:
            print(
                f"[{datetime.now()}] wrote {len(predictions)} records to postgres"
            )
    except Exception as e:
        if isinstance(e, psycopg2.errors.UniqueViolation): pass
        else:
            if verbose:
                print(
                    f"[{datetime.now()}] failed to write {len(predictions)} records to postgres with error {e}"
                )
            return 1
    return 0


def hourly_features_to_postgres(features: pd.DataFrame,
                                engine: "sqlalchemy.engine.Engine",
                                chunksize: int = CHUNCKSIZE,
                                verbose: bool = False) -> int:
    """Load hourly features to postgres database."""
    if verbose:
        print(
            f"[{datetime.now()}] writing {len(features)} records to postgres")
    try:
        features.to_sql(
            "hourly_features",
            engine,
            if_exists="append",
            index=False,
            chunksize=chunksize,
            method="multi",
        )
        if verbose:
            print(
                f"[{datetime.now()}] wrote {len(features)} records to postgres"
            )
    except Exception as e:
        if isinstance(e, psycopg2.errors.UniqueViolation): pass
        else:
            if verbose:
                print(
                    f"[{datetime.now()}] failed to write {len(features)} records to postgres with error {e}"
                )
            return 1
    return 0


def hourly_data_to_postgres(hourly_data_fp: str,
                            engine: "sqlalchemy.engine.Engine",
                            chunksize: int = CHUNCKSIZE,
                            verbose: bool = False) -> int:
    """Load hourly data to postgres database."""
    # get timestamp from file name
    timestamp = datetime.strptime(
        Path(hourly_data_fp).stem.split("_")[1].split(".")[0], "%Y%m%d%H")
    if verbose:
        print(f"[{datetime.now()}] loading {hourly_data_fp} to postgres")
    if verbose: print(f"[{datetime.now()}] file timestamp is {timestamp}")
    try:
        names = download_file(fp=hourly_data_fp, timestamp=timestamp)
        if verbose: print(f"[{datetime.now()}] downloaded {hourly_data_fp}")
        local_fp, blob_name = names
        df = pd.read_csv(local_fp, sep="|", encoding="latin-1", header=None)
        if verbose: print(f"[{datetime.now()}] read {local_fp}")
    except ValueError as e:
        write_failure_to_postgres((timestamp, hourly_data_fp),
                                  "hourly_data_failures", engine=engine)
        print(e)
        return 1
    df.rename(
        columns={
            0: "date",
            1: "time",
            2: "aqsid",
            3: "name",
            4: "unk",
            5: "measurement",
            6: "units",
            7: "value",
            8: "source"
        }, inplace=True)
    # join date and time columns to make a timestamp column
    df["timestamp"] = pd.to_datetime(df["date"] + " " + df["time"],
                                     format="%m/%d/%y %H:%M")
    df.drop(["date", "time"], axis=1, inplace=True)
    try:
        df.to_sql(
            "hourly_data",
            engine,
            if_exists="append",
            index=False,
            chunksize=chunksize,
            method="multi",
        )
        if verbose: print(f"wrote {hourly_data_fp} to postgres")
    except Exception as e:
        if isinstance(e, psycopg2.errors.UniqueViolation): pass
        else:
            if verbose: print(e)
            return 1
    return 0


def load_hourly_data(
    engine: "sqlalchemy.engine.base.Engine",
    table_name: str = "hourly_data",
    features: List[str] = ["PM2.5"],
    start_time: str = "2020-01-01",
    end_time: str = "2024-01-01",
    aqsid_filter: Union[List[str], None] = None,
    verbose: bool = False,
) -> pd.DataFrame:
    """Load the dataframe from the database."""
    measurements = ",".join(f"'{m}'" for m in features)
    aqsid_filter = ",".join(f"'{a}'" for a in aqsid_filter) if aqsid_filter is not None else None
    aqsid_override = f"AND d.aqsid IN ({aqsid_filter})" if aqsid_filter is not None else ""
    query = f"SELECT d.aqsid, d.timestamp, d.measurement, d.value FROM {table_name} AS d WHERE d.timestamp >= '{start_time}' AND d.timestamp < '{end_time}' AND d.measurement IN ({measurements}) {aqsid_override} GROUP BY d.aqsid, d.timestamp, d.measurement, d.value ORDER BY d.timestamp;"

    if verbose: print(f"[{datetime.now()}] executing query: {query}")

    # execute the query
    with engine.connect() as conn:
        query_start = time.time()
        df = pd.read_sql_query(text(query), conn)
        query_end = time.time()
        if verbose:
            print(
                f"[{datetime.now()}] query executed in {query_end - query_start:.2f} seconds"
            )

    if verbose: print(f"[{datetime.now()}] dataframe shape: {df.shape}")

    return df


def load_hourly_feature(
    engine: "sqlalchemy.engine.base.Engine",
    table_name: str = "hourly_features",
    feature: str = "PM2.5",
    start_time: str = "2020-01-01",
    end_time: str = "2024-01-01",
    h3_indices_filter: Union[List[str], None] = None,
    aqsids_filter: Union[List[str], None] = None,
    verbose: bool = False,
) -> pd.DataFrame:
    """Load the dataframe from the database."""
    # build filter for measurements
    measurement_filter = f"'{feature}'"
    # build filter for aqsid (station ids)
    aqsids_filter = ",".join(f"'{a}'" for a in aqsids_filter) if aqsids_filter is not None else None
    aqsids_filter_override = f" AND d.aqsid IN ({aqsids_filter})" if aqsids_filter is not None else ""
    # build filter for h3 index
    h3_indices_filter = ",".join(f"'{a}'" for a in h3_indices_filter) if h3_indices_filter is not None else None
    h3_indices_filter_override = f" AND d.h3_index IN ({h3_indices_filter})" if h3_indices_filter is not None else ""
    
    # build the query
    query = f"SELECT d.h3_index, d.timestamp, d.value FROM {table_name} AS d WHERE d.timestamp >= '{start_time}' AND d.timestamp < '{end_time}' AND d.measurement IN ({measurement_filter}){aqsids_filter_override}{h3_indices_filter_override} GROUP BY d.h3_index, d.timestamp, d.measurement, d.value ORDER BY d.timestamp;"

    if verbose: print(f"[{datetime.now()}] executing query: {query}")

    # execute the query
    with engine.connect() as conn:
        query_start = time.time()
        df = pd.read_sql_query(text(query), conn)
        query_end = time.time()
        if verbose:
            print(
                f"[{datetime.now()}] query executed in {query_end - query_start:.2f} seconds"
            )

    if verbose: print(f"[{datetime.now()}] dataframe shape: {df.shape}")

    return df



import time
from datetime import datetime, timedelta
from pathlib import Path
from typing import Union, List

import pandas as pd
import psycopg2
from sqlalchemy.sql import text

from aq_utilities.config.data import CHUNCKSIZE
from aq_utilities.data.remote import download_file
from aq_utilities.data.failures.psql import write_failure_to_postgres


def stations_info_to_postgres(stations_info_fp: str,
                              engine: "sqlalchemy.engine.Engine",
                              chunksize: int = CHUNCKSIZE,
                              verbose: bool = False) -> int:
    """Load station info data to postgres database."""
    # get timestamp from the directory name
    timestamp_component = Path(stations_info_fp).parent.stem
    if timestamp_component == "today":
        timestamp = datetime.utcnow()
        timestamp = timestamp.replace(hour=0, minute=0, second=0,
                                      microsecond=0)
    elif timestamp_component == "yesterday":
        timestamp = datetime.utcnow() - timedelta(days=1)
        timestamp = timestamp.replace(hour=0, minute=0, second=0,
                                      microsecond=0)
    else:
        timestamp = datetime.strptime(timestamp_component, "%Y%m%d")

    if verbose:
        print(f"[{datetime.now()}] loading {stations_info_fp} to postgres")

    try:
        names = download_file(fp=stations_info_fp, timestamp=timestamp)
        if verbose: print(f"[{datetime.now()}] downloaded {stations_info_fp}")
        local_fp, blob_name = names
        if verbose: print(f"[{datetime.now()}] reading {local_fp}")
        df = pd.read_csv(local_fp, sep="|", encoding="latin-1")
    except Exception as e:
        write_failure_to_postgres((timestamp, stations_info_fp),
                                  "stations_info_failures", engine=engine)
        print(e)
        return 1

    if verbose: print(f"[{datetime.now()}] read {local_fp}")
    # merge the "StateAbbrevation" and "StateAbbreviation" columns into a single column called state_abbreviation
    if "StateAbbrevation" in df and "StateAbbreviation" not in df:
        df.rename(columns={"StateAbbrevation": "state_abbreviation"}, inplace=True)
    elif "StateAbbreviation" in df:
        df.rename(columns={"StateAbbreviation": "state_abbreviation"}, inplace=True)
    else:
        df["state_abbreviation"] = df["StateAbbrevation"].combine_first(df["StateAbbreviation"])
        df.drop(columns=["StateAbbrevation", "StateAbbreviation"], inplace=True)
    columns_to_rename = {
        "AQSID": "aqsid",
        "FullAQSID": "full_aqsid",
        "StationID": "station_id",
        "Parameter": "parameter",
        "MonitorType": "monitor_type",
        "SiteCode": "site_code",
        "SiteName": "site_name",
        "Status": "status",
        "AgencyID": "agency_id",
        "AgencyName": "agency_name",
        "EPARegion": "epa_region",
        "Latitude": "latitude",
        "Longitude": "longitude",
        "Elevation": "elevation",
        "GMTOffset": "gmt_offset",
        "CountryFIPS": "country_fips",
        "CBSA_ID": "cbsa_id",
        "CBSA_Name": "cbsa_name",
        "StateAQSCode": "state_aqs_code",
        "CountyAQSCode": "county_aqs_code",
        "CountyName": "county_name"
    }
    df.rename(columns=columns_to_rename,
            inplace=True)
    df.status = df.status.str.upper()
    # add a column with todays date or date of the file
    df["timestamp"] = timestamp

    try:
        df.to_sql(
            "stations_info",
            engine,
            if_exists="append",
            index=False,
            chunksize=chunksize,
            method="multi",
        )
        if verbose:
            print(f"[{datetime.now()}] wrote {stations_info_fp} to postgres")
    except Exception as e:
        if isinstance(e, psycopg2.errors.UniqueViolation): pass
        else:
            if verbose: print(e)
            return 1
    return 0


def load_stations_info(
    engine: "sqlalchemy.engine.base.Engine",
    table_name: str = "stations_info",
    query_date: str = "2022-01-01",
    aqsid: Union[List[str], None] = None,
    verbose: bool = False,
) -> pd.DataFrame:
    # get the station information for these stations from the database
    aqsid = ",".join(f"'{s}'" for s in aqsid) if aqsid is not None else None
    aqsid_override = f"AND d.aqsid IN ({aqsid})" if aqsid is not None else ""
    query = f"SELECT d.aqsid, d.latitude, d.longitude, d.elevation FROM {table_name} AS d WHERE d.timestamp = '{query_date}' {aqsid_override} GROUP BY d.aqsid, d.latitude, d.longitude, d.elevation ORDER BY d.aqsid;"

    if verbose: print(f"[{datetime.now()}] executing query: {query}")

    with engine.connect() as conn:
        query_start = time.time()
        df_stations = pd.read_sql_query(text(query), conn)
        query_end = time.time()
        if verbose:
            print(
                f"[{datetime.now()}] query executed in {query_end - query_start:.2f} seconds"
            )

    if verbose:
        print(f"[{datetime.now()}] dataframe shape: {df_stations.shape}")

    return df_stations



from datetime import datetime
from typing import Union, Callable, List, Tuple

import pandas as pd


def filter_aqsids(
        stations_info: pd.DataFrame,
        remove_aqsid: Union[List[str], None] = ["000000000"],
        aqsid_col_name: str = "aqsid",
        verbose: bool = False,
    ) -> pd.DataFrame:
    """Filter stations based on aqsid."""
    if verbose:
        print(f"[{datetime.now()}] Filtering from {len(stations_info)} stations.")

    if remove_aqsid is not None:
        if verbose:
            print(f"[{datetime.now()}] Removing stations with aqsid in {remove_aqsid}.")
        stations_info = stations_info[~stations_info[aqsid_col_name].isin(remove_aqsid)]

    if verbose:
        print(f"[{datetime.now()}] Returning {len(stations_info)} stations.")  

    return stations_info


def filter_lat_lon(
        stations_info: pd.DataFrame,
        remove_lat_lon: Union[List[Tuple[float, float]], None] = (0, 0),
        lat_col_name: str = "latitude",
        lon_col_name: str = "longitude",
        verbose: bool = False,
    ) -> pd.DataFrame:
    """Filter stations based on lat-lon locations."""
    if verbose:
        print(f"[{datetime.now()}] Filtering from {len(stations_info)} stations.")

    if remove_lat_lon is not None:
        if verbose:
            print(f"[{datetime.now()}] Removing stations with lat-long paris in {remove_lat_lon}.")
        stations_info = stations_info[
            ~stations_info[[lat_col_name, lon_col_name]].apply(
                lambda x: tuple(x) in remove_lat_lon, axis=1
            )
        ]
    
    if verbose:
        print(f"[{datetime.now()}] Returning {len(stations_info)} stations.")  

    return stations_info


def round_station_lat_lon(
        stations_info: pd.DataFrame,
        round_lat_lon: Union[int, None] = 2,
        lat_col_name: str = "latitude",
        lon_col_name: str = "longitude",
        verbose: bool = False,
    ) -> pd.DataFrame:
    """Filter stations based on aqsid."""
    if verbose:
        print(f"[{datetime.now()}] Filtering from {len(stations_info)} stations.")

    if round_lat_lon is not None:
        if verbose:
            print(f"[{datetime.now()}] Rounding lat and lon to {round_lat_lon} decimal places.")
        stations_info = stations_info.round({lat_col_name: round_lat_lon, lon_col_name: round_lat_lon})
    
    if verbose:
        print(f"[{datetime.now()}] Returning {len(stations_info)} stations.")  

    return stations_info


def remove_duplicate_lat_lon(
        df: pd.DataFrame,
        lat_col_name: str = "latitude",
        lon_col_name: str = "longitude",
        verbose: bool = False,
    ) -> pd.DataFrame:
    """Remove duplicate lat-lon pairs."""
    if verbose:
        print(f"[{datetime.now()}] Removing duplicates from {len(df)} rows.")

    df = df.drop_duplicates(subset=[lat_col_name, lon_col_name])

    if verbose:
        print(f"[{datetime.now()}] Returning {len(df)} rows.")

    return df


def remove_duplicate_aqsid(
        df: pd.DataFrame,
        aqsid_col_name: str = "aqsid",
        verbose: bool = False,
    ) -> pd.DataFrame:
    """Remove duplicate aqsid pairs."""
    if verbose:
        print(f"[{datetime.now()}] Removing duplicates from {len(df)} rows.")

    df = df.drop_duplicates(subset=[aqsid_col_name])

    if verbose:
        print(f"[{datetime.now()}] Returning {len(df)} rows.")

    return df


def apply_filters(
        df: pd.DataFrame,
        filters: List[Callable],
        verbose: bool = False,
    ) -> pd.DataFrame:
    """Apply a list of filters to a dataframe."""
    for f in filters:
        if verbose:
            print(f"[{datetime.now()}] Applying filter {f.__name__}.")
        df = f(df, verbose=verbose)
    return df



from datetime import datetime
from pathlib import Path
from urllib.request import urlopen
from typing import Tuple, Union


def download_file(
    fp: str,
    timestamp: datetime,
) -> Union[Tuple[str, str], None]:
    """Download the file at fp and save it to disk.
    
    Args:
        fp: the url of the file to download
        timestamp: the timestamp to use for the blob name

    Returns:
        name, blob_name: the name of the file on disk and the name of the blob
    """
    # download the file to local storage
    with urlopen(fp) as response:
        # read the file
        try:
            data = response.read()
        except Exception as e:
            raise ValueError(f"failed to read file at {fp} with error {e}")
        # determine the file name
        name = Path(fp).name
        # construct the blob name using the timestamp
        blob_name = f"{timestamp.year}/{timestamp.month}/{timestamp.day}/{name}"
        # save the data to disk
        with open(name, "wb") as f:
            f.write(data)
        return name, blob_name



from datetime import datetime
from typing import Union

from sqlalchemy import text


def get_max_timestamp(table_name: str, timestamp_col_name: str = "timestamp",
                      engine: "sqlalchemy.engine.Engine" = None,
                      verbose: bool = False) -> Union[datetime, None]:
    """Get max timestamp from postgres table."""
    start_date = None
    try:
        with engine.connect() as connection:
            if verbose:
                print(
                    f"[{datetime.now()}] getting max timestamp from {table_name}"
                )
            result = connection.execute(
                text(f"SELECT MAX({timestamp_col_name}) FROM {table_name};"))
            for row in result:
                start_date = row[0]
                if verbose:
                    print(
                        f"[{datetime.now()}] got max timestamp of {start_date} from {table_name}"
                    )
                break
    except Exception as e:
        print(f"failed to get max timestamp from {table_name} with error {e}")
        return None

    if isinstance(start_date, str):
        start_date = datetime.strptime(start_date, "%Y-%m-%d %H:%M:%S")
    return start_date


def get_min_timestamp(table_name: str, timestamp_col_name: str = "timestamp",
                      engine: "sqlalchemy.engine.Engine" = None,
                      verbose: bool = False) -> Union[datetime, None]:
    """Get min timestamp from postgres table."""
    start_date = None
    try:
        with engine.connect() as connection:
            if verbose:
                print(
                    f"[{datetime.now()}] getting min timestamp from {table_name}"
                )
            result = connection.execute(
                text(f"SELECT MIN({timestamp_col_name}) FROM {table_name};"))
            for row in result:
                start_date = row[0]
                if verbose:
                    print(
                        f"[{datetime.now()}] got min timestamp of {start_date} from {table_name}"
                    )
                break
    except Exception as e:
        print(f"failed to get min timestamp from {table_name} with error {e}")
        return None

    if isinstance(start_date, str):
        start_date = datetime.strptime(start_date, "%Y-%m-%d %H:%M:%S")
    return start_date



import time
from datetime import datetime, timedelta
from pathlib import Path
from typing import List, Union

import pandas as pd
import numpy as np
from sqlalchemy import text
import psycopg2

from aq_utilities.config import CHUNCKSIZE
from aq_utilities.data.remote import download_file
from aq_utilities.data.failures.psql import write_failure_to_postgres


def daily_stations_to_postgres(daily_stations_fp: str,
                               engine: "sqlalchemy.engine.Engine",
                               chunksize: int = CHUNCKSIZE,
                               verbose: bool = False) -> int:
    """Load daily stations data to postgres database."""
    # get timestamp from the directory name
    timestamp_component = Path(daily_stations_fp).parent.stem
    if verbose:
        print(f"[{datetime.now()}] loading {daily_stations_fp} to postgres")
    if verbose: print(f"[{datetime.now()}] file timestamp is {timestamp}")
    if timestamp_component == "today":
        timestamp = datetime.utcnow()
        timestamp = timestamp.replace(hour=0, minute=0, second=0,
                                      microsecond=0)
    elif timestamp_component == "yesterday":
        timestamp = datetime.utcnow() - timedelta(days=1)
        timestamp = timestamp.replace(hour=0, minute=0, second=0,
                                      microsecond=0)
    else:
        timestamp = datetime.strptime(timestamp_component, "%Y%m%d")
    try:
        names = download_file(fp=daily_stations_fp, timestamp=timestamp)
        if verbose: print(f"[{datetime.now()}] downloaded {daily_stations_fp}")
        local_fp, blob_name = names
        df = df = pd.read_csv(local_fp, sep="|", encoding="latin-1",
                              header=None)
        if verbose: print(f"[{datetime.now()}] read {local_fp}")
    except ValueError as e:
        write_failure_to_postgres((timestamp, daily_stations_fp),
                                  "daily_stations_failures", engine=engine)
        print(e)
        return 1
    df = df[[0, 1, 4, 8, 9]]
    # obtain list of unique values for each column based on StationID
    df = df.groupby(0).agg(lambda x: list(sorted(np.unique(x)))
                           if x.name == 1 else x.iloc[0])
    df.rename(
        columns={
            0: "aqsid",
            1: "parameters",
            4: "status",
            8: "latitude",
            9: "longitude"
        }, inplace=True)
    # cast status to upper case
    df.status = df.status.str.upper()
    # add a column with todays date or date of the file
    df["timestamp"] = timestamp
    df = df.rename_axis("aqsid").reset_index()

    try:
        df.to_sql(
            "daily_stations",
            engine,
            if_exists="append",
            index=False,
            chunksize=chunksize,
            method="multi",
        )
        if verbose: print(f"wrote {daily_stations_fp} to postgres")
    except Exception as e:
        if isinstance(e, psycopg2.errors.UniqueViolation): pass
        else:
            print(e)
            return 1
    return 0


def load_daily_stations(
    engine: "sqlalchemy.engine.base.Engine",
    table_name: str = "daily_stations",
    query_date: str = "2022-01-01",
    aqsid: Union[List[str], None] = None,
    measurements: Union[List[str], None] = None,
    verbose: bool = False,
) -> pd.DataFrame:
    # get the station information for these stations from the database
    aqsid = ",".join(f"'{s}'" for s in aqsid) if aqsid is not None else None
    aqsid_override = f" AND d.aqsid IN ({aqsid})" if aqsid is not None else ""
    # make a measurements filter if measurements are provided
    measurements = ",".join(f"'{m}'" for m in measurements) if measurements is not None else None
    measurements_override = f" AND d.parameters && ARRAY[{measurements}]" if measurements is not None else ""
    
    query = f"SELECT d.aqsid, d.latitude, d.longitude FROM {table_name} AS d WHERE d.timestamp = '{query_date}'{aqsid_override}{measurements_override} GROUP BY d.aqsid, d.latitude, d.longitude ORDER BY d.aqsid;"

    if verbose: print(f"[{datetime.now()}] executing query: {query}")

    with engine.connect() as conn:
        query_start = time.time()
        df_stations = pd.read_sql_query(text(query), conn)
        query_end = time.time()
        if verbose:
            print(
                f"[{datetime.now()}] query executed in {query_end - query_start:.2f} seconds"
            )

    if verbose:
        print(f"[{datetime.now()}] dataframe shape: {df_stations.shape}")

    return df_stations



from datetime import datetime
from typing import Union, Callable, List, Tuple

import h3
import numpy as np
import pandas as pd


def measurements_to_aqsid(
    df: pd.DataFrame,
    stations_info: pd.DataFrame,
    start_time: str = "2020-01-01",
    end_time: str = "2024-01-01",
    time_step: str = "1H",
    aggregation_method: Union[Callable, str] = lambda x: np.mean(x[x>0]) if len(x[x>0]) > 0 else np.nan,
    nan_value: Union[float, int] = np.nan,
    time_closed_interval: bool = False,
    verbose: bool = False,
) -> pd.DataFrame:
    """Take the raw DataFrame from the database and process it into a fixed-station, fixed-index format."""
    # make a datetime index for the full range
    date_range = pd.date_range(
        start=start_time, end=end_time, freq=time_step,
        inclusive="left" if time_closed_interval else "both")

    if verbose: print(f"[{datetime.now()}] processing dataframe")

    # we would like df to be indexed on time and have columns for each station
    df = df[["aqsid", "timestamp",
             "value"]].groupby(["timestamp", "aqsid"]).aggregate({
                 "value":
                 aggregation_method
             }).reset_index().pivot(index="timestamp", columns="aqsid",
                                    values="value").resample(time_step).agg(aggregation_method).reindex(date_range)
    # undo the pivot so that we have a column for timestamp, a column for aqsid, and a column for value
    df = df.unstack().reset_index().rename(columns={
        "level_1": "timestamp",
        0: "value"
    })
    # join the lat, lon, elevation information to the main dataframe
    stations_info = stations_info.groupby("aqsid").first()
    df = df.join(stations_info, on="aqsid", how="inner")

    if verbose: print(f"[{datetime.now()}] dataframe shape: {df.shape}")

    # fill missing values with negative 1
    df = df.fillna(nan_value)

    return df


def measurements_to_h3(
    df: pd.DataFrame,
    start_time: str,
    end_time: str,
    aggregation_method: Union[Callable, str],
    min_h3_resolution: int,
    leaf_h3_resolution: int,
    include_root_node: bool = True,
    time_closed_interval: bool = False,
    verbose: bool = False,
) -> np.ndarray:
    """Compute h3 indexed measurements from aqsid index measurements."""
    # ensure that the columns exist
    assert "timestamp" in df.columns
    assert "latitude" in df.columns
    assert "longitude" in df.columns
    assert "aqsid" in df.columns
    assert "value" in df.columns

    # get the features from start time to end time
    if time_closed_interval:
        df = df[(df.timestamp >= start_time)
                & (df.timestamp <= end_time)].copy(deep=True)
    else:
        df = df[(df.timestamp >= start_time)
                & (df.timestamp < end_time)].copy(deep=True)
    df.reset_index(inplace=True)

    if verbose: print(f"[{datetime.now()}] processing feature")

    # we have one station that is at the exact same location as another station, so we need to drop one of them
    df = df.groupby(["timestamp", "latitude", "longitude"]).aggregate({
        "aqsid":
        "first",
        "value":
        aggregation_method
    }).reset_index()
    if verbose: print(f"[{datetime.now()}] dataframe shape: {df.shape}")

    # map the h3 index at the leaf resolution to the station
    # pivot the dataframe so that we have a column for each timestamp and a row for each h3 index
    df = df.pivot(index=["latitude", "longitude"], columns="timestamp",
                  values="value").reset_index()
    if verbose:
        print(f"[{datetime.now()}] pivoted dataframe shape: {df.shape}")

    if verbose:
        print(f"[{datetime.now()}] processing resolution {leaf_h3_resolution}")
    df["h3_index"] = df.apply(
        lambda x: h3.geo_to_h3(x.latitude, x.longitude, leaf_h3_resolution),
        axis=1)
    df.drop(columns=["latitude", "longitude"], inplace=True)

    h3_indexed_features = [
    ]  # tuples of (h3_index, value_t1, ..., value_tn), when concatenated will be shape (num_nodes, num_timestamps)

    # iterate through the h3 indices between the leaf resolution and the coarsest resolution
    for next_h3_resolution in range(leaf_h3_resolution - 1,
                                    min_h3_resolution - 2, -1):
        # ensure that the h3_index is the first column
        col_names = df.columns.tolist()
        col_names.remove("h3_index")
        # sort the columns by timestamp
        col_names.sort()
        col_names.insert(0, "h3_index")
        df = df[col_names]
        # add values for the current resolution
        h3_indexed_features.extend(df.to_numpy())

        if next_h3_resolution < min_h3_resolution: break
        if verbose:
            print(
                f"[{datetime.now()}] processing resolution {next_h3_resolution}"
            )
        # get the h3 index for each station at the next_resolution
        df["next_h3_index"] = df.apply(
            lambda x: h3.h3_to_parent(x.h3_index, next_h3_resolution), axis=1)
        # group by the next h3 index
        df.drop(columns=["h3_index"], inplace=True)
        df = df.groupby("next_h3_index").aggregate(
            aggregation_method).reset_index()
        # rename the h3 index to the current h3 index
        df = df.rename(columns={"next_h3_index": "h3_index"})

    # add another parent node for the entire graph
    if include_root_node:
        if verbose: print(f"[{datetime.now()}] adding root node")
        root_node_id = "root"
        df.drop(columns=["h3_index"], inplace=True)
        h3_indexed_features.append(
            np.concatenate((np.array([root_node_id]),
                            df.apply(aggregation_method).values.T)))

    return h3_indexed_features


def determine_leaf_h3_resolution(
    df: pd.DataFrame,
    min_h3_resolution: int = 0,
    max_h3_resolution: int = 12,
    verbose: bool = False,
) -> int:
    """Determine the leaf resolution that will give us unique hexagons for each station."""
    # we have one station that is at the exact same location as another station, so we need to drop one of them
    assert "latitude" in df.columns and "longitude" in df.columns, "df must have latitude and longitude columns"
    assert "aqsid" in df.columns, "df must have aqsid column"

    df = df.groupby(["latitude", "longitude"]).aggregate({
        "aqsid": "first"
    }).reset_index()

    # find the MAX_H3_RESOLUTION that will give us unique hexagons for each station
    leaf_h3_resolution: int = min_h3_resolution
    if verbose: print(f"[{datetime.now()}] determining leaf resolution")
    for i in range(min_h3_resolution, max_h3_resolution):
        if verbose: print(f"[{datetime.now()}] trying resolution {i}")
        if len(df["aqsid"].values) == len(
                df.apply(lambda x: h3.geo_to_h3(x.latitude, x.longitude, i),
                         axis=1).unique()):
            leaf_h3_resolution = i
            if verbose: print(f"[{datetime.now()}] found resolution {i}")
            break

    return leaf_h3_resolution


def aqsid_to_h3(
    df: pd.DataFrame,
    min_h3_resolution: int = 0,
    leaf_h3_resolution: int = 9,
    include_root_node: bool = True,
    as_df: bool = True,
    verbose: bool = False,
) -> List[Tuple[Union[str, None], str, int]]:
    """Process the edges and edge features for the graph."""
    assert "latitude" in df.columns and "longitude" in df.columns, "df must have latitude and longitude columns"
    if verbose:
        print(f"[{datetime.now()}] processing resolution {leaf_h3_resolution}")
        print(
            f"[{datetime.now()}] after resolution {leaf_h3_resolution}, we have {len(df)} nodes"
        )
    # map the h3 index at the leaf resolution to the station
    df["h3_index"] = df.apply(
        lambda x: h3.geo_to_h3(x.latitude, x.longitude, leaf_h3_resolution),
        axis=1)
    # store as a dataframe, adding new rows for each resolution
    node_df = df[["aqsid", "h3_index"]].copy(deep=True)

    # the aqsid is not meaningful except at the leaf resolution
    df["aqsid"] = None

    # iterate through the h3 indices between the leaf resolution and the coarsest resolution
    for next_h3_resolution in range(leaf_h3_resolution - 1,
                                    min_h3_resolution - 2, -1):
        if next_h3_resolution < min_h3_resolution: break
        if verbose:
            print(
                f"[{datetime.now()}] processing resolution {next_h3_resolution}"
            )
        # get the h3 index for each station at the next_resolution
        df["next_h3_index"] = df.apply(
            lambda x: h3.h3_to_parent(x.h3_index, next_h3_resolution), axis=1)
        # group by the next h3 index
        df = df.groupby("next_h3_index").aggregate({
            "aqsid": "first"
        }).reset_index()  # we don't need to aggregate the values
        # rename the h3 index to the current h3 index
        df = df.rename(columns={"next_h3_index": "h3_index"})
        # add these new rows to the node_df
        node_df = pd.concat([node_df, df[["aqsid", "h3_index"]]])
        if verbose:
            print(
                f"[{datetime.now()}] after resolution {next_h3_resolution}, we have {len(node_df)} nodes"
            )

    # add another parent node for the entire graph if needed
    if include_root_node:
        if verbose: print(f"[{datetime.now()}] adding root node")
        root_node_id = "root"
        node_df = pd.concat([
            node_df,
            pd.DataFrame({
                "aqsid": [None],
                "h3_index": [root_node_id]
            })
        ])

    # the node id is the index
    node_df["node_id"] = range(len(node_df))
    if verbose: print(f"[{datetime.now()}] processed {len(node_df)} nodes")

    # return as a dataframe
    if as_df:
        return node_df
    # return as a list of tuples
    return node_df.to_numpy()


def ensure_feature_axis(
        df: pd.DataFrame,
        h3_indices: List[str],
        timestamps: Union[List[Union[str, datetime]], pd.DatetimeIndex],
        timestamp_col_name: str = "timestamp",
        h3_index_col_name: str = "h3_index",
        values_col_name: str = "value",
        verbose: bool = False,
) -> pd.DataFrame:
    """Ensure the return from remote has the correct columns and rows."""
    # ensure that the columns are the correct type
    df[timestamp_col_name] = pd.to_datetime(df[timestamp_col_name])
    
    if verbose: print(f"[{datetime.now()}] initial dataframe shape: {df.shape}")
    
    df = df.pivot(index=timestamp_col_name, columns=h3_index_col_name, values=values_col_name).reindex(index=timestamps, columns=h3_indices)
    df = df.reset_index()
    df = df.rename(columns={"index": timestamp_col_name})
    df = df.melt(id_vars=timestamp_col_name, value_vars=h3_indices, var_name=h3_index_col_name, value_name=values_col_name)
    
    if verbose: print(f"[{datetime.now()}] final dataframe shape: {df.shape}")
    
    return df
